import os
import re

from glob import glob


class Processes(object):
    RE_ESCAPES  = re.compile(r'[\\\$\\ \'"]') # match letters \, $, ', " and space
    def __init__(self):
        self.processes = {}

        # process inspection time
        for process in glob('/proc/*'):
            if not os.path.isdir(process):
                continue

            pid = os.path.split(process)[1]
            try:
                pid = int(pid)
            except ValueError:
                continue

            # find parent pid
            with open(os.path.join(process, 'status')) as f:
                status = f.read()
                parent = int(re.search('PPid:\t(\\d+)', status).group(1))

            # find cwd
            try:
                cwd = os.readlink(os.path.join(process, 'cwd'))
            except OSError:
                cwd = None

            # find command
            with open(os.path.join(process, 'cmdline')) as f:
                cmd = self.make_safe_cmd(f.read())

            self.processes[pid] = dict(parent=parent, pid=pid, cwd=cwd, cmd=cmd, children=[])

        # fixup children pointers
        for k, v in self.processes.items():
            parent = v['parent']
            if parent in self.processes:
                self.processes[v['parent']]['children'].append(k)

    def __getitem__(self, item):
        return self.processes.__getitem__(item)

    def get(self, item):
        return self.processes.get(item)

    def make_safe_cmd(self, cmd):
        cmd = [self.make_safe_arg(each) for each in cmd.split('\0')]
        return ' '.join(cmd)

    def make_safe_arg(self, arg):
        def escape(m):
            return '\\' + m.group(0)
        return self.RE_ESCAPES.sub(escape, arg)

if __name__ == '__main__':
    p = Processes()
    pid = os.getpid()
    while 1:
        print p[pid]
        pid = p[pid]['parent']
        if not pid:
            break



def run():
    import argparse
    import logging
    import os
    import sys
    import stat

    from . import commands
    from . import process

    default_socket = '/tmp/muxpy-%s/default' % os.getuid()

    p = argparse.ArgumentParser(prog='muxpy', description='manage tmux sessions')
    p.add_argument('-S', '--socket', default=default_socket, help='the tmux socket. defaults to ' + default_socket)
    p.add_argument('-v', '--verbose', default=0, action='count', help='set verbose mode. repeat up to 3 times')

    subp = p.add_subparsers(dest='command', title='available subcommands')

    help = {
        'create': dict(
            help='create a new profile from an existing session',
            description='create a new session'),
        'edit': dict(
            help='edit a saved session',
            description='edit a profile manually'),
        'start': dict(
            description='start a saved session',
            help='start a saved profile'),
        'attach': dict(
            description='attach to a socket',
            help='attach to an existing tmux session'),
        'kill': dict(
            description='kill a socket', 
            help='kill an existing tmux session'),
    }

    def add_profile(p):
        p.add_argument('profile', help='the profile name')
        p.add_argument('-f', '--format', choices=['json'], default='json', help='profile format')
            
    helpp = subp.add_parser('help')
    helpp.add_argument('target', help='help on what?', nargs='?')

    createp = subp.add_parser('create', **help['create'])
    add_profile(createp)

    editp = subp.add_parser('edit', **help['edit'])
    add_profile(editp)

    startp = subp.add_parser('start', **help['start'])
    add_profile(startp)

    subp.add_parser('attach', **help['attach'])
    subp.add_parser('kill', **help['kill'])

    parsed = p.parse_args(sys.argv[1:])

    logging.basicConfig(format='')
    logger = logging.getLogger()
    levels = {0: logging.ERROR, 1: logging.WARNING, 2: logging.INFO, 3: logging.DEBUG}
    logger.setLevel(levels.get(parsed.verbose, 3))
#     logger.warning('logging set to %s' % logging.getLevelName(levels.get(parsed.verbose, 3)))

    if parsed.command == 'help':
        args = ['--help']
        if parsed.target:
            args.insert(0, parsed.target)

        p.parse_args(args)
        sys.exit(1)

    def err_out(msg):
        logger.error(msg)
        p.print_help()
        sys.exit(1)

    if parsed.command in  ['create', 'attach']:
        try:
            s = os.stat(parsed.socket)
        except OSError:
            err_out('%s does not exist' % parsed.socket)

        if not stat.S_ISSOCK(s.st_mode):
            err_out('%s is not a socket' % parsed.socket)

    try:
        if parsed.command == 'start':
            commands.start(parsed)
        elif parsed.command == 'create':
            commands.create(parsed)
        elif parsed.command == 'edit':
            commands.edit(parsed)
        elif parsed.command == 'attach':
            commands.attach(parsed)
        elif parsed.command == 'kill':
            commands.kill(parsed)
    except process.TmuxExecError as e:
        logger.error(e.message)

    sys.exit(0)



import json

import logging
logger = logging.getLogger()

def parse(file_format, data):
    if file_format == 'json':
        logger.info('reading %d bytes as json', len(data))
        return json.loads(data)
    else:
        raise NotImplementedError('there is no support yet for %s' % file_format)


def format(file_format, data):
    if file_format == 'json':
        json_data = json.dumps(data, indent=4)
        logger.info('dumping %d bytes as json', len(json_data))
        return json_data
    else:
        raise NotImplementedError('there is no support yet for %s' % file_format)



import os
from subprocess import Popen, PIPE

import logging
logger = logging.getLogger()


class TmuxExecError(RuntimeError):
    pass


class TmuxExecutor(object):
    def __init__(self, socket=None):
        if not socket:
            socket = '/tmp/tmux-%s/default' % os.getuid()

        self.socket = socket
        pname, sname = os.path.split(socket)

        if pname and not os.path.isdir(pname):
            os.makedirs(pname)

        logger.debug('initializing %s with socket:%s', self.__class__, socket)
        self.cmd = ['/usr/bin/env', 'tmux', '-S', socket]

    def __call__(self, *command):
        cmd = self.cmd[:]
        cmd.append('--')
        cmd.extend(command)
        return self._execute(cmd)

    def _execute(self, cmd):
        logger.info('executing tmux: %s', ' '.join(cmd))
        p = Popen(cmd, stdout=PIPE, stderr=PIPE)
        out, err = p.communicate()
        if p.returncode != 0:
            raise TmuxExecError('error executing %s: %s' % (' '.join(cmd), err))

        return out


if __name__ == '__main__':
    tmux = TmuxExecutor('/tmp/tmux-1000/default')
    print tmux('list-sessions'),
    print tmux('list-windows', '-t', 'muxpy'),
    print tmux('list-panes', '-a', '-F', "#{session_name}\t#{window_number}").split('\t'),



import os
p = os.path


def get_profile_path(profile, format):
    user = p.expanduser('~')
    path = p.join(user, '.muxpy', 'profiles', "%s.%s" % (profile, format))
    return p.abspath(path)


if __name__ == '__main__':
    print get_profile_path('testprof', 'json')



# -*- coding: utf-8 -*-
from distutils.core import setup
import re


def get_version():
    init_py = open('muxpy/__init__.py').read()
    metadata = dict(re.findall("__([a-z]+)__ = '([^']+)'", init_py))
    return metadata['version']


setup(
    name='Muxpy',
    version=get_version(),
    author='Bjørnar Snoksrud',
    author_email='bjornar@snoksrud.no',
    packages=['muxpy'],
    scripts=['bin/muxpy'],
    license='ICS',
    description='A tmux session handler',
    long_description=open('README').read(),
    classifiers=[
        'Programming Language :: Python :: 2.7',
    ],
    url='https://github.com/bjornars/muxpy',
)



import logging
import os
import sys

from collections import OrderedDict
from itertools import count

from . import profile, parser, formats, backend

logger = logging.getLogger()

def edit(p):
    prof = profile.get_profile_path(p.profile, p.format)
    cmd = '/etc/alternatives/editor "%s"' % prof

    logger.info('executing %s', cmd)
    os.system(cmd)


def start(p):
    prof = profile.get_profile_path(p.profile, p.format)
    if not os.path.isfile(prof):
        print >> sys.stderr, 'Cannot find profile %s' % p.profile
        sys.exit(1)

    tmux = backend.Backend(socket=p.socket)

    with open(prof) as f:
        data = formats.parse(p.format, f.read())

    for session in data:
        tmux.create_session(session=session['name'], window=session['windows'][0]['name'])

        for window in session['windows'][1:]:
            tmux.new_window(window['name'])

        for window in session['windows']:
            tmux.select_window(window['name'])

            # make the required number of panes
            for index, pane in enumerate(window['panes']):
                if index > 0:
                    tmux.new_pane()

                if pane and 'cwd' in pane:
                    tmux.send_keys('cd %s\n' % pane['cwd'])

                if pane and 'cmd' in pane:
                    tmux.send_keys('%s\n' % pane['cmd'])

            tmux.select_layout(window['layout'])

    attach(p)


def create(p):
    prof = profile.get_profile_path(p.profile, p.format)
    pname, fname = os.path.split(prof)
    if not os.path.isdir(pname):
        os.makedirs(pname)

    data = []

    tmux = parser.TmuxParser(p.socket)
    sessions = tmux.list_sessions()
    panes = tmux.get_panes()

    for session in sessions:
        session_data = OrderedDict(name=session.name)
        windows = tmux.list_windows(session.name)

        unique_windows = []
        for window in windows:
            # make sure the window name is unique, or else tmux will weep loudly
            if not window['name'] in unique_windows:
                unique_windows.append(window['name'])
            else:
                for x in count():
                    new_window = "%s_%d" % (window['name'], x)
                    if new_window not in unique_windows:
                        break
                unique_windows.append(new_window)
                window['name'] = new_window

            window['panes'] = panes[session['name']][window['number']]

        session_data['windows'] = windows
        data.append(session_data)

    formatted = formats.format(p.format, data)

    with open(prof, 'w') as f:
        f.write(formatted)

    print 'made profile, %d sessions with total %d windows' % (len(data), sum(len(x['windows']) for x in data))

def attach(p):
    cmd = 'tmux -u2 -S %s attach' % p.socket
    logger.info('executing %s', cmd)
    os.system(cmd)

def kill(p):
    tmux = backend.Backend(socket=p.socket)
    tmux.kill_server()
    os.unlink(p.socket)
    print 'killed tmux at %s' % p.socket



from .process import TmuxExecutor


class Backend(object):
    def __init__(self, socket):
        self.socket = socket
        self.tmux = TmuxExecutor(socket)

    def list_windows(self, session):
        return self.tmux('list-windows', '-t', session)

    def list_panes(self):
        return self.tmux('list-panes', '-a', '-F', '#{session_name}\t#{window_index}\t#{pane_index}\t#{pane_pid}')

    def list_sessions(self):
        return self.tmux('list-sessions')

    def create_session(self, session=None, window=None):
        args = ['new-session', '-d', '-x', '200', '-y', '200']
        if session:
            args.extend(['-s', session])
        if window:
            args.extend(['-n', window])

        return self.tmux(*args)

    def new_window(self, name):
        return self.tmux('new-window', '-d', '-n', name)

    def select_layout(self, layout):
        return self.tmux('select-layout', layout)

    def select_window(self, name):
        return self.tmux('select-window', '-t', name)

    def new_pane(self):
        self.tmux('split-window', '-h', '-p', '100')

    def send_keys(self, keys):
        self.tmux('send-keys', '%s' % keys)

    def kill_server(self):
        self.tmux('kill-server')

if __name__ == '__main__':
    tmux = Backend('/tmp/tmux-1000/default')
    print tmux.list_panes()



import re
from collections import defaultdict
from itertools import groupby
from operator import itemgetter

from .backend import Backend
from .ps import Processes

RE_SESSIONS = re.compile(r"^(?P<name>.*?): (?P<windows>\d*)")
RE_WINDOWS = re.compile(r"^(?P<number>\d*?): (?P<name>.*) \[\d+x\d+\] \[layout (?P<layout>.*?)$")


class ddict(dict):
    def __getattr__(self, item):
        return self.get(item, False)

    def __missing__(self, item):
        return False


class TmuxParser(object):
    def __init__(self, socket=None, backend=Backend):
        self.backend = backend(socket)

    def intify(self, dict_, keys):
        for k, v in list(dict_.iteritems()):
            if k in keys:
                dict_[k] = int(v)

    def list_sessions(self):
        sessions = filter(None, self.backend.list_sessions().split('\n'))

        matches = [RE_SESSIONS.search(each) for each in sessions]
        groups = [ddict(m.groupdict()) for m in matches if m]

        for each in groups:
            self.intify(each, 'windows')

        return groups

    def list_windows(self, session):
        windows = filter(None, self.backend.list_windows(session).split('\n'))
        matches = [RE_WINDOWS.search(each) for each in windows]
        groups = [ddict(m.groupdict()) for m in matches if m]

        for each in groups:
            self.intify(each, 'number')
            layout = each['layout']
            if layout.endswith(' (active)'):
                layout = layout[:-9]

            each['layout'] = layout[:-1]

        return groups

    def get_panes(self, get_process_info=True):
        panes = filter(None, self.backend.list_panes().split('\n'))
        panes = map(lambda x: x.split('\t'), panes)
        panes.sort()
        pane_dict = defaultdict(dict)

        if get_process_info:
            processes = Processes()

        def get_pane_info(pid):
            process = get_process_info and processes.get(pid)

            if process and len(process['children']) == 1:
                child = processes[process['children'][0]]
                return dict(cwd=child['cwd'], cmd=child['cmd'])
            else:
                return {}

        for session, windows in groupby(panes, itemgetter(0)):
            windows = [x[1:] for x in windows]

            for window, panes in groupby(windows, itemgetter(0)):
                pane_dict[session][int(window)] = [get_pane_info(int(each[2])) for each in panes]

        return dict(pane_dict)


if __name__ == '__main__':
    tmux = TmuxParser(socket='/tmp/tmux-1000/default')
    print tmux.list_sessions()
    print tmux.list_windows('muxpy')
